/* This program uses the Arduino Nano (3.0, ATmega328) over a Serial link, to control a Delay-Line (implemented with an external fast FIFO).
 *
 * Copyright (C) Richard Neill 2012, <arduino at REMOVE.ME.richardneill.org>. This program is Free Software. You can 
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, 
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied. 
 * For the details, please see: http://www.gnu.org/licenses/gpl.html 
 *
 * TODO: Possible enhancements: add ability to control other pins, or capture analog values.
 *
 * The Communications Protocol is very simple. Each side transfers one line at a time, ending with "\n" ("\r\n" is also accepted by Arduino.).
 * Commands are case-insensitve. Arduino response ends with "\n". The commands marked with "*" below are the most useful ones.
 * [The Arduino should respond within 1 second in all cases, (measure is slowest), usually much faster.]
 * When compiled with debugging enabled, extra messages begin "DEBUG: ".
 *
 *     PC SENDS         ARDUINO REPLIES                                            ARDUINO DOES
 *     --------         ---------------                                            ------------
 *   * HELLO            DELAY LINE                                                 Nothing: just say hello
 *   * CONFIG           OSC_MHZ: 25, DELAY RANGE: [AAA, BBB] NS, OR 0 TO BYPASS.   Prints configuration.
 *     HELP             COMMANDS: HELLO, CONFIG, HELP, BLINK, BYPASS, ...          This lists the (important) commands: "HELLO; CONFIG; SETDELAY INT_NS; GETDELAY; GETVDELAY; BYPASS; INSERT; MEASURE; COUNT INT_MS; SLOWSWL; HELP; VERSION; ABOUT; UPTIME; BLINK; RESET"
 *     VERSION          VERSION 1.0                                                Protocol version. (Overkill?)
 *     ABOUT            FIFO CONTROL. COPYRIGHT XXX, XXX. FREE SOFTWARE GNU GPL V3+   Prints information about this.
 *     UPTIME           UP: SSS SECONDS                                            Print uptime in seconds. Wraps after ~ 50 days
 *     BLINK            BLINK                                                      Blink the diagnostic LED.
 *     BYPASS           NOFIFO                                                     Bypasses the FIFO. Data is not delayed at all.
 *     INSERT           FIFO                                                       Inserts the FIFO into the signal path. Data is delayed (this doesn't change the value of the fifo delay).
 *   * GETDELAY         NNN                                                        Current delay in NS. Returns integer number of ns; 0 if bypassed; -1 if uninitialised and not bypassed.
 *     GETVDELAY        WWW, XXX, YYY, ZZZ  #NS, +/-NS, TICKS, MHZ                 Get verbose delay value. Values are respectively: delay(ns), error(ns), ticks, clock(MHz). If FIFO disabled, all values are 0. (-1 for uninitialised).
 *   * SETDELAY DDD     OK                                                         Set new delay value to DDD nanoseconds (coerced by clock ticks). Implicitly inserts the fifo (or bypasses for 0).
 *     SETDELAY EEE     INVALID VALUE: 'EEE'. RANGE IS: [XXX,XXX] NS               Error setting new delay. Value is out of range.
 *     MEASURE          TTT                                                        Measure the delay value (not exact), and print it. Value in ns (resolution ~ 10us). -1 if timed out (at 0.5 second). Doesn't affect the main delay-line.
 *     COUNT    TTT     NNN                                                        Count the number of falling-edges arriving, for TTT ms. The count is exact (for f <~ 2.5 MHz); ms is accurate to about 0.2%.
 *     SLOWSWL          SLOWSWL: 60...   (multiple lines, 1 Hz for 1 minute)       Test mode for the PCB: clocks the FIFO at 40/25 kHz (1000x slower than normal). Manual entry of data with SWitch,Led. Ends after 1 minute.
 *     RESET            RESET                                                      Resets the Arduino in software. This also clears the FIFO.  (A *Hardware* reset can be done with the serial port DTR line.)
 *     TEST             TEST                                                       For experimentation: not normally used.          #Not included in help string.
 *     ANALOG0          AAA                                                        Get analog pin0 value.                           #Not included in help string.
 *     ""               ""                                                         Empty string: does nothing. Returns newline.     #Not included in help string.
 *     MAIL             undocumented                                               Try it and see.                                  #Not included in help string.
 *     WHY              undocumented                                               Try it and see.                                  #Not included in help string.
 *     [anything]       INVALID COMMAND: 'xxx'. TRY 'HELP'.                        Erroneous command. Returns error message.
 *
 *
 *
 * The Arduino is connected as follows:
 *
 *     Pin (Port,Bit)     Named                          Description
 *     --------------     --------------                 -----------
 *     13  (PortB,5)      LED_DEBUG                      This is the onboard diagnostic LED.
 *     12  (PortB,4)      FIFO_DATA_IN4                  FIFO input (Data bit 4). Arduino output.
 *     11  (PortB,3)      FIFO_DATA_OUT4                 FIFO output (Data bit 4). Arduino input.
 *     10  (PortB,2)      FIFO_nRST                      FIFO nReset. Both nRRST and nWRST. (direct connection)
 *     9   (PortB,1)      LED_FIFO                       Output opposite of MUX_nA_B. High to show FIFO is inserted, Low (LED off) for Bypassed.
 *     8   (PortB,0)      LED_PREFILL                    High during FIFO prefill, low otherwise.
 *     7   (PortD,7)      OSC_OE                         Output enable for oscillator. High to enable, Low to pause.
 *     6   (PortD,6)      FIFO_nWE                       FIFO nWrite_Enable.  (via Multiplexer)
 *     5   (PortD,5)      COUNTER_IN                     FIFO output (data bit 0). Arduino Async 16-bit counter input. (Hardware special pin).                             .
 *     4   (PortD,4)      MUX_nA_B                       Multiplexer nA/B. Low for Fifo, High for Bypass (and preloading of fifo).
 *     3   (PortD,3)      FIFO_nRE                       FIFO nRead_Enable.  (via Multiplexer)
 *     2   (PortD,2)      FIFO_RWCLK                     FIFO RClock and WClock (via Multiplexer)
 *
 */


/* Definitions */
#define DEBUG          0           /* Enable verbosity if 1, else 0 */
#define BAUD_RATE      115200      /* USB-Serial port speed */
#define FIFO_MIN       128         /* Min and max depth of the AL422 FIFO. Note Min depth > 0: see datasheet for why. */
#define FIFO_MAX       393216
#define XTAL_OSC_MHZ   40          /* This is the speed of the attached oscillator for external fifo, NOT the Arduino */
#define EXT_OSC_MHZ    25          /* The oscillator module can be removed, and replaced with an external clock input. */
#define MAX_LEN        128         /* Max size of string (overflows won't crash the program) */
#define COUNT_MAX_MS   1000000     /* Max counting time in ms. (NB possible serial-communication timeout) */
#define DEBUG_PREFIX   "DEBUG: "
#define VERSION_STR    "1.0"        /* The wire-protocol version, not the version of the firmware. */
#define AUTHOR         "RICHARD NEILL"
#define DATE           "2012"
#define LICENSE        "FREE SOFTWARE, GNU GPL V3+"

#define USE_ONBOARD_OSC 0          /* 1 for onboard 40 MHz osc. module; 0 for external inbound clock, 25 MHz. */
#if (USE_ONBOARD_OSC == 1)
        #define OSC_MHZ    XTAL_OSC_MHZ
        #define OSC_OE_EN  HIGH
#else
        #define OSC_MHZ    EXT_OSC_MHZ
        #define OSC_OE_EN  LOW
#endif
#define CLOCK_NS   (1000/OSC_MHZ)
#define MEASURE_TMOUT_US  ((CLOCK_NS * FIFO_MAX) / 100)   /* Timeout for measurement (with safety factor of 5) */

#define ON             true
#define OFF            false
#define INSERT         false
#define BYPASS         true
#define LED_DEBUG      13
#define FIFO_DATA_IN4  12
#define FIFO_DATA_IN4_PORT  PORTB
#define FIFO_DATA_IN4_BIT  4
#define FIFO_DATA_OUT4 11
#define FIFO_nRST      10
#define LED_FIFO       9
#define LED_PREFILL    8
#define OSC_OE         7
#define FIFO_nWE       6
#define COUNTER_IN     5
#define MUX_nA_B       4
#define FIFO_nRE       3
#define FIFO_RWCLK     2
#define FIFO_RWCLK_PORT PORTD
#define FIFO_RWCLK_BIT 2


/* Macros */
#define STRINGIFY(s)          STRINGIFY2(s)    /* convert a constant eg 4 to the string "4" in the preprocesor */
#define STRINGIFY2(s)         #s               /* BUT this can't convert eg (3+5) to "8"; instead it emits the very unhelpful "(3+5)" */
#define debugSerial(...)      if (DEBUG) { Serial.print(DEBUG_PREFIX); Serial.print ( __VA_ARGS__); Serial.print("\n"); };
#define dbg_led(state)        digitalWrite (LED_DEBUG, state);   /* Compiles to fewer bytes as a macro than a function call! */
#define set_FIFO_RWCLK()      FIFO_RWCLK_PORT |=  (0x01 << FIFO_RWCLK_BIT)   /* 10x faster than digitalWrite */
#define clr_FIFO_RWCLK()      FIFO_RWCLK_PORT &= ~(0x01 << FIFO_RWCLK_BIT)
#define set_FIFO_DATA_IN4()   FIFO_DATA_IN4_PORT |= (0x01 << FIFO_DATA_IN4_BIT)


/* Globals */
char inputString[MAX_LEN]="";       //Command from serial port
long fifo_ns=-1;                    //Number of fifo nanoseconds
long fifo_ticks=-1;                 //Number of fifo ticks
long fifo_jitter_ns=-1;             //Jitter in the delay (half a clock cycle).
boolean fifo_enabled=false;         //Is the fifo enabled or bypassed?

unsigned long overflow_count;         //Globals, for the timed counter.
unsigned int timer_ticks, timer_period;
volatile unsigned long timer_counts;  //These ones are changed by the Interrupt Service Routines,
volatile boolean counter_ready;       //so must be declared volatile.


/* Reset in (Software) */
void (* sw_reset) (void) = 0;       //declare reset function @ address 0


/* Debugging LED (pin 13) */
void dbg_led_toggle(){
        digitalWrite (LED_DEBUG, (!digitalRead (LED_DEBUG)) );
}


/* Trim string, convert to long. *sptr points to the start of the string, value is the answer (returned by reference).
 * Retval is boolean: true means the entire string represented a number. */
boolean param2long ( char *sptr, long *value){
        boolean on_lhs = true;
        char *endptr;

        for (int i = 0; i < strlen(sptr); i++){    //rtrim() spaces after first non-space.
                if (on_lhs){                       //No need to ltrim() spaces, strtol() does that.
                        if (sptr[i] != ' '){
                                on_lhs = false;
                        }
                }else if (sptr[i] == ' '){
                        sptr[i] = '\0';
                        break;
                }
        }
        //debugSerial (sptr);

        *value = strtol(sptr, &endptr, 10);   //convert string to long int. (base 10)
        if (*sptr != NULL && *endptr == NULL){ //conversion OK; the entire string is valid.
                //debugSerial (*value);
                return (true);
        }else{
                 //debugSerial ("STRTOL() FAILED");
                 return (false);
        }
}


/* This function is for experimentation. Undocumented, and subject to change. */
void do_test(){
       /* Test performance of digitalWrite vs direct port access */
       Serial.print ("TEST START\n");
       for (long i=0 ; i< FIFO_MAX; i++){
                digitalWrite (LED_DEBUG, LOW);      //slower
       }
       Serial.print ("TEST MIDDLE\n");
       for (long i=0 ; i< FIFO_MAX; i++){
                PORTB |= (0x01 << (LED_DEBUG - 8)); //10x faster
       }
       Serial.print ("TEST END\n");
}


/* Manually clock the FIFO at ~ 40 or 25 kHz. The main signal bypasses it, but testing signals
 * (using a switch + LED on a spare FIFO bit) are now delayed 1000x longer than normal,
 * i.e. delay is 0-10 seconds, rather than 0-10 ms. This test mode runs for 60 seconds. */
void do_slowswl(){
       boolean fifo_enabled_save;
       fifo_enabled_save = fifo_enabled;  //Save state.
       bypass_fifo (BYPASS);
       int t1_us = CLOCK_NS/2;        /* Half the clock period, x1000 */
       int t2_us = CLOCK_NS - t1_us;  /* t2 and t1 add exactly, even if CLOCK_NS is odd */
       for (int i=60; i > 0; i--){
                Serial.print ("SLOWSWL: ");   Serial.print(i);   Serial.print ("...\n");
                for (long j=0; j < 40000; j++){
                        set_FIFO_RWCLK();
                        delayMicroseconds(t1_us);
                        clr_FIFO_RWCLK();
                        delayMicroseconds(t2_us);
                }
        }
	if (fifo_enabled_save){
	        bypass_fifo (INSERT);
	}
}


/* Blocking readline from serial port, similar to fgets().
 * Read until newline, then return uppercase(data) (excluding the newline) in inputString.
 * If string length exceeds MAX_LEN, the string will be emptied, and reads will continue. */
void readSerial(){
        int i=0;
        char incomingByte;
        while (1){
                if (Serial.available() > 0) {  //If there is data, read a byte.
                        incomingByte = toupper(Serial.read());
                        if (incomingByte == '\r') {        //Ignore "\r"
                                continue;
                        }
                        if (incomingByte == '\n') {        //End of string? Exclude the "\n".
                                break;
                        }
                        inputString[i++] = incomingByte;   //Append
                        if (i >= MAX_LEN){  //Too long?
                                inputString[0] = incomingByte;
                                i=0;
                        }
                }
        }
        inputString[i]='\0';
}


/* Enable or Bypass the FIFO. */
void bypass_fifo(boolean state){
        /* We could, if desired, set OSC_OE inactive (!OSC_OE_EN) when bypassed, to stop the 40 MHz oscillator.*/
        /* This would save power, and some RFI, but NB oscillator startup time 10 ms */
        if (state == BYPASS){    //Bypass it.
                digitalWrite(MUX_nA_B, BYPASS);
                digitalWrite(LED_FIFO, !BYPASS);
                fifo_enabled = false;
        }else{                   //Enable it.
                digitalWrite(MUX_nA_B, INSERT);
                digitalWrite(LED_FIFO, !INSERT);
                fifo_enabled= true;
        }
}


/* Reset the FIFO at boot (and leave it bypassed). */
/* After reset, FIFO could just be directly inserted, but should call configure_fifo(). */
void reset_fifo (){
        debugSerial ("RESETTING FIFO");

        bypass_fifo (BYPASS);                 //Put the FIFO into bypass/program mode. Manually clock it.

        digitalWrite(FIFO_RWCLK, LOW);        //Take (both) the clocks low.

        digitalWrite(FIFO_nRE, HIGH);         //Take ReadEnable, WriteEnable high (so no writes during reset).
        digitalWrite(FIFO_nWE, HIGH);

        digitalWrite(FIFO_nRST, LOW);         //Reset it. Take Reset low, and wait 1ms
        delay(1);
        digitalWrite(FIFO_RWCLK, HIGH);       //Rising edge on Clock, to make the reset actually happen.
        digitalWrite(FIFO_nRST, HIGH);        //Take the FIFO out of reset.

        digitalWrite(FIFO_nRE, LOW);          //Prepare to activate: make ReadEnable and WriteEnable low.
        digitalWrite(FIFO_nWE, LOW);
}


/* Configure the FIFO. This implicitly enables it. fifo_ticks is known to be valid, in range, non-zero. */
/* We aren't trying to put valid data into the FIFO, just to set up the right distance between the read and write pointers */
void configure_fifo (long fifo_ticks){
        long i;
        debugSerial ("PREFILLING FIFO...");
        debugSerial (fifo_ticks);

        digitalWrite(LED_PREFILL, HIGH);      //Indicate we are prefilling it.
        bypass_fifo (BYPASS);                 //Put the FIFO into bypass/program mode. Manually clock it.

        digitalWrite(FIFO_RWCLK, LOW);        //Take (both) the clocks low.

        digitalWrite(FIFO_nRE, HIGH);         //Take ReadEnable, WriteEnable high (so no writes during what follows).
        digitalWrite(FIFO_nWE, HIGH);

        digitalWrite(FIFO_nRST, LOW);         //Reset it. Take Reset low, and wait 1ms
        delay(1);
        digitalWrite(FIFO_RWCLK, HIGH);       //Rising edge on Clock, to make the reset actually happen.
        digitalWrite(FIFO_nRST, HIGH);        //Take the FIFO out of reset.

        digitalWrite(FIFO_nWE, LOW);          //WriteEnable low: enable writes on clock rising-edge
        for (i=0; i < fifo_ticks; i++){       //Clock the FIFO n times.
                clr_FIFO_RWCLK();             //These macros are 10x faster than digitalWrite() !
                set_FIFO_RWCLK();
        }

        digitalWrite(FIFO_nRE, LOW);          //Prepare to activate: make ReadEnable and WriteEnable low.
        digitalWrite(FIFO_nWE, LOW);          //(nWE is already low anyway)

        bypass_fifo (INSERT);                 //Re-enable FIFO: insert it into the signal path and connect to the master clock.
        digitalWrite(LED_PREFILL, LOW);       //Done.
        debugSerial ("FINISHED PREFILLING FIFO.");
}


/* Measure the fifo delay, against our own crystal clock. Return a number of ns, or 0 if bypassed, or -1 if timeout.
 * This can be run live: it doesn't affect the delay-line. */
long measure_fifo(){
        unsigned long then, now, timeout, delta;
        boolean timedout = true;

        if (!fifo_enabled){                    //If the FIFO is bypassed, then the real delay actually is 0.
                return (0);		       //ALSO, it's not being clocked, so we couldn't measure the fifo delay anyway.
        }

        /* Ideally, we'd like to disable interrupts [i.e. noInterrupts() ] between here and the end of the while() loop. They do cause jitter.
         * Unfortunately, interrupts are used by micros() to handle 8-bit timer overflows, so we can't. */
        then =  micros();                      //Get start time in microseconds. (NB this can wrap at 2^32)
        timeout = then + MEASURE_TMOUT_US;     //Will timeout if we get here.

        set_FIFO_DATA_IN4();                   //Send rising edge to FIFO input. (We were initialised low). (Use faster macro, for efficiency).

        while (micros() < timeout){            //Loop, polling for the received pulse. Both micros() and timeout wrap together.
              if (digitalRead( FIFO_DATA_OUT4)){
                      now = micros();          //Get end time.
                      timedout = false;        //NB this loop takes 8us to execute: that is the limit on resolution. Could do better with timer-counter.
                      break;
              }
        }
        if (timedout){
                now = micros();
        }else{
                now -= 12;                      //The 12 is measured with FIFO_DATA_IN1 connected to FIFO_DATA_OUT1. This is the minimum; it varies.
        }
        delta = (now - then);                   //Difference (us). Wraps are ok.
        debugSerial (then); debugSerial (now); debugSerial (delta);

        digitalWrite( FIFO_DATA_IN4, LOW);     //Clean up...send low into the fifo for a while.
        delay (delta/800);                     //Wait slightly longer than delta. (NB conversion us to ms)

        if (timedout){                         //If we timed out, result is -1.
                return (-1);
        }else{                                 //Otherwise, return the measurement (converted to ns).
                return ( delta * 1000 );
        }
}


/* Count pulses on TMR1 input, using the 16-bit asynchronous TMR1, for duration_ms. */
unsigned long count_pulses (long duration_ms){  /* 1 < duration < COUNT_MAX_MS */
         /* There are 3 timers in the ATmega328. Timer0 (8-bit), Timer1 (16-bit), Timer2 (8-bit).
          * Timer0 is not used: the Arduino uses it for the system functions like delay() and millis().
          * Timer1 is used asynchronously, clocked by falling edges on pin T1 (aka COUNTER_IN or D5).
          *   No prescaler is used. Interrupt on overflow; this increments overflow_count.
          * Timer2 is used with the system-clock, Prescaler of 128, and a count of 125 => exactly 1000us.
          *    On matching, the interrupt checks whether we have reached target ms.
          * There is a very helpful example here: http://www.gammon.com.au/forum/?id=11504

	  * Note that the circuit (ab)uses the FIFO to buffer the input signal; so the fifo must be inserted
	  * into the path (then configured short, and allowed to fill) for this to work. This is a bug. */

	boolean fifo_enabled_save;     //Initial state of fifo_enabled.
        counter_ready = false;         //ISR on counter 2 sets this when done.
        timer_period = duration_ms;    //nunmber of 1 mS counts to do.
        timer_ticks = 0;               //count interrupts of timer2
        overflow_count = 0;            //no overflows yet

	fifo_enabled_save = fifo_enabled;  //Save state. (also fifo_ticks is preserved)
	configure_fifo (FIFO_MIN);	   //Set the fifo to shortest delay and insert it (work around the circuit layout bug).
	delayMicroseconds((FIFO_MIN/OSC_MHZ) +1); //Allow it to fill (this is only a few us).

        TCCR1A = TCCR1B = 0;           //Initialise both Timer 1 and Timer 2  (also stops their clocks)
        TCCR2A = TCCR2B = 0;
        TCNT1 = 0;                     //Zero both counters
        TCNT2 = 0;

        TIFR1 = _BV (TOV1);            //Clear interrupt flag (it might be set already)
        TIMSK1 = _BV (TOIE1);          //Enable interrupt on overflow of timer 1: TIMER1_OVF_vect

        TCCR2A = _BV (WGM21);          //Timer2: Comparison mode
        OCR2A  = 124;                  //Compare with 124, means count 125 times.

        TIFR2 = _BV (OCF2A);           //Clear interrupt flag (it might be set already)
        TIMSK2 = _BV (OCIE2A);         //Enable interrupt on timer 2 match: TIMER2_COMPA_vect

        GTCCR = _BV (PSRASY);          //Reset prescalers. [One common register-bit affects all 3 timers {0,1,2} ! Will make millis() glitch.]

        TCCR2B =  _BV (CS20) | _BV (CS22); //Start Timer 2, using a prescaler of 128.

        TCCR1B =  _BV (CS11) | _BV (CS12); //Start Timer 1, using external falling-edge clock on T1 (pin D5).  (Add "| _BV(CS10)" for rising edge.)

        long startm = micros();
        while (!counter_ready){        //Busy-wait, till counter is ready.
                  /* As the counters/timers roll over, the interrupts fire. */
        }
        long endm = micros();
        float freq = ((float)timer_counts *  1000.0) / timer_period; //Calculate frequency
        debugSerial (endm-startm);
        debugSerial (freq);

	if (fifo_enabled_save){       // Restore original state.
		if (fifo_ticks != -1){
			configure_fifo (fifo_ticks);
		}
	}else{
		bypass_fifo (BYPASS);
	}

        return (timer_counts);  /* Result: number of pulses counted. */
}


/* Interrupt: handle timer 1 overflow. */
ISR (TIMER1_OVF_vect){
        overflow_count++;     //number of Counter1 overflows
}


/* Interrupt: timer2 comparison: triggered every 1ms while counting is active. */
ISR (TIMER2_COMPA_vect) {
        unsigned int timer1_counter_value = TCNT1;  //Get current value of Counter1.

        if (++timer_ticks < timer_period){  //Have we reached end of timing period?
                return;  //...not yet
        }

        //End of gate time, measurement ready.
        TIMSK1 = TIMSK2 = 0;    //disable Timer1 and Timer2 Interrupt
        TCCR1A = TCCR1B = 0;    //stop timer 1
        TCCR2A = TCCR2B = 0;    //stop timer 2

        //Calculate total count
        timer_counts = (overflow_count << 16) + timer1_counter_value;  //each overflow is 65536 more
        counter_ready = true;              // set global flag for end count period
}


/* Setup code: run at power-on */
void setup (){
        Serial.begin(BAUD_RATE);               //Configure serial port
        Serial.flush();
        debugSerial ("DEBUGGING ENABLED");

        if (F_CPU != 16000000){                //Check CPU freq. If we are using a board with a
                 while (1){                    //non-16MHz crystal, the code in count_pulses() will be wrong.
                         Serial.print("Error: CPU frequency is ");   Serial.print(F_CPU);   Serial.print(", but expect 16 MHz.\n");
                         delay (100);
                 }
        }

        pinMode (LED_DEBUG, OUTPUT);           //Set up outputs.
        pinMode (MUX_nA_B, OUTPUT);
        pinMode (FIFO_nWE, OUTPUT);
        pinMode (FIFO_nRE, OUTPUT);
        pinMode (FIFO_nRST, OUTPUT);
        pinMode (FIFO_RWCLK, OUTPUT);
        pinMode (LED_FIFO, OUTPUT);
        pinMode (LED_PREFILL, OUTPUT);
        pinMode (FIFO_DATA_IN4, OUTPUT);
        pinMode (OSC_OE, OUTPUT);

        pinMode (FIFO_DATA_OUT4, INPUT);       //Set up inputs.
        pinMode (COUNTER_IN, INPUT);

        pinMode (A0, INPUT);                   //These pins are unused.
        pinMode (A1, INPUT);                   //Set them explicitly to inputs (they should be anyway)
        pinMode (A2, INPUT);                   //and then enable the weak (20k) pullup resistors.
        pinMode (A3, INPUT);
        pinMode (A4, INPUT);
        pinMode (A5, INPUT);
        pinMode (A6, INPUT);
        pinMode (A7, INPUT);
        digitalWrite (A0, HIGH);
        digitalWrite (A1, HIGH);
        digitalWrite (A2, HIGH);
        digitalWrite (A3, HIGH);
        digitalWrite (A4, HIGH);
        digitalWrite (A5, HIGH);
        digitalWrite (A6, HIGH);
        digitalWrite (A7, HIGH);

        reset_fifo();                          //Reset and Bypass the fifo at boot.

        digitalWrite (FIFO_DATA_IN4, LOW);     //Initial values. (most pins were set by reset_fifo() ).
        digitalWrite (OSC_OE, OSC_OE_EN);
        digitalWrite (LED_PREFILL, LOW);

        for (char i=0; i < 10; i++){           //Flash Debug LED 5x at boot, to show we just (re)started,
                dbg_led_toggle();              //then stay on.
                delay(50);
        }
        dbg_led (ON);
}






/* Main loop. */
void loop() {

        readSerial();
        debugSerial(inputString);    /* For debugging */

        if (!strcmp(inputString, "HELLO") ){            /* "HELLO" command. Respond with: "DELAY LINE" */
                Serial.print ("DELAY LINE\n");

        }else if (!strcmp (inputString, "CONFIG") ){    /* "CONFIG" command. Respond with: version, oscillator speed, and delay range. */
                Serial.print ("OSC_MHZ: ");   Serial.print(OSC_MHZ);   Serial.print(", DELAY RANGE: [");   Serial.print(CLOCK_NS * FIFO_MIN);  Serial.print(", ");  Serial.print(CLOCK_NS * FIFO_MAX);   Serial.print("] NS, OR 0 TO BYPASS.\n");

        }else if (!strcmp(inputString, "HELP") ){        /* "HELP" command. Respond with: list of commands */
                Serial.print ("COMMANDS: HELLO; CONFIG; SETDELAY INT_NS; GETDELAY; GETVDELAY; BYPASS; INSERT; MEASURE; COUNT INT_MS; SLOWSWL; HELP; VERSION; ABOUT; UPTIME; BLINK; RESET\n");

        }else if (!strcmp(inputString, "VERSION") ){     /* "VERSION" command. Respond with: "VERSION: XXX" */
                Serial.print ("VERSION: " VERSION_STR "\n");

        }else if (!strcmp(inputString, "ABOUT") ){       /* "ABOUT" command. Respond with: Author/Copyright info. */
                Serial.print ("FIFO CONTROL. COPYRIGHT " AUTHOR ", "DATE". " LICENSE ".\n");

        }else if (!strcmp(inputString, "UPTIME") ){      /* "UPTIME" command. Respond with: "UP: XXX SECONDS" */
                Serial.print ("UP: ");   Serial.print ( millis()/1000 );   Serial.print (" SECONDS\n");   /* millis() wraps after ~ 50 days. */

        }else if (!strcmp(inputString, "BLINK") ){       /* "BLINK" command. Blink the LED. Respond with: "BLINK". */
                dbg_led_toggle();
                delay(100);
                dbg_led_toggle();
                delay(100);
                Serial.print ("BLINK\n");

        }else if (!strcmp(inputString, "BYPASS") ){     /* "BYPASS" command. Bypass the FIFO. Respond with: "NOFIFO". */
                bypass_fifo(BYPASS);
                Serial.print ("NOFIFO\n");

        }else if (!strcmp(inputString, "INSERT") ){     /* "INSERT" command. Insert the FIFO. Respond with: "FIFO". */
                bypass_fifo(INSERT);
                Serial.print ("FIFO\n");

        }else if (!strcmp(inputString, "GETDELAY") ){   /* "GETDELAY" command. Return the delay, in nanoseconds. 0 if bypassed; -1 if uninitialised (but not bypassed). */
                if (fifo_enabled){
                        Serial.print (fifo_ns);   Serial.print("\n");
                }else{
                        Serial.print ("0\n");
                }

        }else if (!strcmp(inputString, "GETVDELAY") ){      /* "GETVDELAY" command. Find current delay value and print it. Also print Ticks value and Oscillator MHz. */
                if (fifo_enabled){
                        Serial.print (fifo_ns);         Serial.print (", ");
                        Serial.print (fifo_jitter_ns);  Serial.print (", ");
                        Serial.print (fifo_ticks);      Serial.print (", ");
                        Serial.print (OSC_MHZ);         Serial.print ("  #NS, +/-NS, TICKS, MHZ\n");
                }else{
                        Serial.print ("0, 0, 0, 0  #NS, +/-NS, TICKS, MHZ\n");  //All zeros if fifo is bypassed.
                }

        }else if (!strncmp(inputString, "SETDELAY ",strlen("SETDELAY ")) ){     /* "SETDELAY" command. Parse string, set delay, print the result. */
                long value;
                char *sptr = &inputString[strlen("SETDELAY ")]; /* Start at the end of "SETDELAY ", then parse the rest to long */
                boolean ret = param2long ( sptr, &value);

                if ( (ret == false) || ( (value < 0) || (value > 0 && value < (CLOCK_NS * FIFO_MIN)) || (value > (CLOCK_NS * FIFO_MAX)) ) ){ //Invalid, or out of range
                        debugSerial("STRTOL() FAILED, OR DATA OUT OF RANGE.");
                        Serial.print ("INVALID VALUE: '");       Serial.print (sptr);
                        Serial.print ("'. RANGE IS: 0, or [");   Serial.print(CLOCK_NS * FIFO_MIN);   Serial.print (", ");   Serial.print(CLOCK_NS * FIFO_MAX);   Serial.print("] NS.\n");

                }else{                        //Valid, and in range.
                        if ( value == 0){         //Zero. Bypass the fifo. (Can't program a zero delay).
                                bypass_fifo (BYPASS);
                        }
                        else{                    //Non-zero, and within range. Program the fifo.
                                fifo_ns = value;
                                fifo_ticks = fifo_ns / CLOCK_NS;  //integer division.
                                fifo_ns = fifo_ticks * CLOCK_NS;  //...and calculate coerced value.
                                fifo_jitter_ns = CLOCK_NS/2;
                                configure_fifo (fifo_ticks);  //set up the fifo.
                        }
                        Serial.print  ("OK\n");
                }

        }else if (!strcmp(inputString, "MEASURE") ){   /* "MEASURE" command: try to measure the delay of the fifo against our own clock */
                long delay_length = measure_fifo();
                Serial.print (delay_length);  Serial.print ("\n");

        }else if (!strncmp(inputString, "COUNT ",strlen("COUNT ")) ){   /* "COUNT" command. Count falling edges during time-period, print the result. */
                long value;                                             /* Counting uses the 16-bit asynchronous TMR1 counter. */
                char *sptr = &inputString[strlen("COUNT ")]; /* cf setdelay */
                boolean ret = param2long ( sptr, &value);

                if ( (ret == false) || ( (value < 1) || (value > COUNT_MAX_MS) ) ){ //Invalid, or out of range
                        debugSerial("STRTOL() FAILED, OR DATA OUT OF RANGE.");
                        Serial.print ("INVALID VALUE: '");   Serial.print (sptr);  Serial.print ("'. RANGE IS: [1, "); Serial.print(COUNT_MAX_MS);  Serial.print("] MS\n");
                }else{                        //Valid, and in range.
                        unsigned long count = count_pulses (value);  //Count for value ms, return number.
                        Serial.print (count);   Serial.print ("\n");
                }

        }else if (!strcmp(inputString, "SLOWSWL") ){    /* "SLOWSWL" command: for PCB testing */
                do_slowswl();
                Serial.print ("SLOWSWL\n");

        }else if (!strcmp(inputString, "RESET") ){     /* "RESET" command: reset the chip */
                Serial.print ("RESET\n");
                delay(50);
                sw_reset();

        }else if (!strcmp(inputString, "ANALOG0") ){   /* "ANALOG0" command: Get the value from Analog Pin 0 */
                Serial.print (analogRead(A0));         /* NB, the weak pullup is enabled */
                Serial.print("\n");

        }else if (!strcmp(inputString, "TEST") ){     /* "TEST" command: for development */
                do_test();
                Serial.print ("TEST\n");

        }else if (!strcmp(inputString, "WHY") ){      /* "WHY" command. */
                Serial.print ("42\n");

        }else if (!strcmp(inputString, "MAIL") ){     /* "MAIL" command. */
                Serial.print ("EVERY PROGRAM ATTEMPTS TO EXPAND UNTIL IT CAN READ MAIL. THOSE PROGRAMS WHICH CANNOT SO EXPAND ARE REPLACED BY ONES WHICH CAN.\n");

        }else if (!strcmp(inputString, "") ){         /* "". Empty string. Reply with blank line" */
                Serial.print ("\n");

        }else{                                        /* "???" Invalid. Respond with "INVALID COMMAND:..." */
                Serial.print ("INVALID COMMAND: '");   Serial.print (inputString);   Serial.print ("'. TRY 'HELP'.\n");
        }
}

